"""
Daline体系中所有的单元类型定义,unit负责定义接口,frame负责框架和执行。所有业务单元均派生于以下基础单元之一
UnitBase: 单元基类,作为PU,TU等单元父类,不直接作为业务单元父类
ProcessUnit: 常驻业务服务的基类,派生常驻型后台服务
TaskUnit: 一次性或定期任务型脚本基类,做一次性或定期数据处理任务,不直接对外提供服务接口
TradeStrategyUnit: 量化交易策略单元基类,包含策略回测,实盘托管等接口定义
任何业务单元仅支持从预设单元类型中派生,单实例。按规范要求实现业务逻辑并在规定接口中返回数据结果,并要按规范要求编写docstring,代码编写及各类命名遵循编码规范,pylint无报错。
"""
import os
from urllib import response
import oss2
import threading
import time
import redis
import traceback
import requests
import random
# import logging
from functools import wraps
from logging import Logger
from abc import ABC, abstractmethod
from sqlalchemy.orm import Session

def db_session(obj_func):
    """
    自动获取并释放数据库session给需要使用的函数
    """
    @wraps(obj_func)
    def wrap_dbuse_fun(unit_ins,*args, **kwargs):
        with Session(unit_ins.sql_engine) as session:
            unit_ins.set_sql_context_session(session)
            ret = obj_func(unit_ins,*args, **kwargs)
            session.close()
            unit_ins.set_sql_context_session(None)
            return ret
    return wrap_dbuse_fun
class UnitBase(ABC):
    """
    所有单元基类,主要处理日志、接口路由等基础功能

    Attributes:
        _logger (logger): 日志实例
    """

    def __init__(self):
        """
        初始化
        """
        self._logger = None
        self._sql_config = {}
        self._sql_engine = None
        self._sql_context_session = None
        self._redis_pool = None
        self._folder_temp = ''
        self._folder_nfs = ''
        self._folder_nfs_author = ''
        self._folder_http = ''
        self._http_base = ''
        self._folder_oss = ''
        self._oss_bucket = None
        self._request_func = None
        self._request_glb_func = None
        self._reg_request = {}
        self.GLB_REQUEST_START = 5000
        self._reg_glb_request = {}
        self._subscription_func = None
        self._unsubscription_func = None
        self._reg_subscription = {}

    @abstractmethod
    def test(self)->bool:
        """
        各业务单元自定义单元测试部分

        Returns:
            True: 自动测试成功
            False: 自动测试失败
        """
        return True

    @property
    def logger(self)->Logger:
        """
        使用logger实例
        """
        return self._logger
        # return logging.getLogger()

    @logger.setter
    def logger(self,value):
        """
        设置本单元所用logger实例

        Parameters:
            logger (logger): logger实例
        """
        self._logger = value

    @property
    def sql_config(self):
        """
        数据库连接配置
        """
        return self._sql_config
    def set_sql_config(self,config):
        """
        设置数据库连接引擎

        Parameters:
            engine (sql engine): 数据库连接引擎
        """
        self._sql_config = config
    @property
    def sql_connection(self):
        """
        使用sql connection实例
        """
        if self._sql_engine:
            return self._sql_engine.connect()
        else:
            return None
    @property
    def sql_engine(self):
        """
        数据库连接引擎实例
        """
        return self._sql_engine
    def set_sql_engine(self,engine):
        """
        设置数据库连接引擎

        Parameters:
            engine (sql engine): 数据库连接引擎
        """
        self._sql_engine = engine
    def set_sql_context_session(self,session):
        """
        设置上下文session,给db_session装饰的目标函数使用

        Parameters:
            session (Session): 上下文session
        """
        self._sql_context_session = session
    def sql_fetch_session(self)->Session:
        """
        使用sql session实例
        """
        return self._sql_context_session

    @property
    def redis_conn(self):
        """
        数据库连接引擎实例
        """
        if self._redis_pool:
            return redis.Redis(connection_pool=self._redis_pool)
        else:
            return None
    def set_redis_pool(self,redis_pool):
        """
        设置redis连接池

        Parameters:
            redis_pool (redis pool): redis连接池
        """
        self._redis_pool = redis_pool

    @property
    def folder_temp(self)->str:
        """
        获取临时存储目录
        """
        return self._folder_temp
    @folder_temp.setter
    def folder_temp(self,path)->str:
        """
        设置本单元所用临时存储目录,仅做临时缓存使用,仅本单元可见,单元重启后数据可能丢失

        Parameters:
            path (string): 临时存储目录的绝对路径
        """
        self._folder_temp = path

    @property
    def folder_nfs(self)->str:
        """
        获取单元用NFS存储目录
        """
        return self._folder_nfs
    @folder_nfs.setter
    def folder_nfs(self,path)->str:
        """
        设置本单元本版本所用NFS存储目录,永久存储,同名业务单元的多个实例共享,单元重启后数据不丢失

        Parameters:
            path (string): 存储目录的绝对路径
        """
        self._folder_nfs = path
    @property
    def folder_nfs_author(self)->str:
        """
        获取本作者NFS存储目录,各单元存储目录都在本作者目录下
        """
        return self._folder_nfs_author
    @folder_nfs_author.setter
    def folder_nfs_author(self,path)->str:
        """
        设置本作者所用NFS存储目录,永久存储,其名下所有PU及PU各版本目录均在此目录下

        Parameters:
            path (string): 存储目录的绝对路径
        """
        self._folder_nfs_author = path

    def set_http_base(self,base):
        """
        设置本单元所用web存储目录的基础链接地址

        Parameters:
            base (string): 基础链接地址
        """
        self._http_base = base
    @property
    def folder_http(self)->str:
        """
        获取本单元本版本用http存储目录
        """
        return self._folder_http
    @property
    def folder_http_unit(self)->str:
        """
        获取单元用http存储目录
        """
        return os.path.split(self.folder_http)[0]
    @property
    def folder_http_author(self)->str:
        """
        获取本作者用http存储目录
        """
        return os.path.split(self.folder_http_unit)[0]
    @folder_http.setter
    def folder_http(self,path):
        """
        设置本单元所用web存储目录,同OSS存储定位相似,永久存储,同名业务单元的多个实例共享,单元重启后数据不丢失,有URL链接地址,可通过http链接公开访问到

        Parameters:
            path (string): 存储目录的绝对路径
        """
        self._folder_http = path
    def http_link(self,path)->str:
        """
        获取http存储中path对应公开访问的http链接

        Parameters:
            path (string): 包含folder_http的绝对路径
        """
        return path.replace(self._folder_http,self._http_base)

    def init_oss(self,access_key_id,access_key_secret,endpoint,bucket_name,folder):
        """
        设置OSS参数

        Parameters:
            access_key_id (string): oss key
            access_key_secret (string): oss secret
            endpoint (string): oss endpoint
            bucket_name (string): oss 区块名
            folder (string): 所使用的目录
        """
        self._oss_bucket = oss2.Bucket(
            oss2.Auth(access_key_id,access_key_secret),
            endpoint,
            bucket_name)
        self.oss_set_folder(folder)
    def oss_get_folder(self)->str:
        """
        个人文件下创建文件夹
        Parameters:
            folder (string): 文件夹名称
        """
        return self._folder_oss
    def oss_set_folder(self, folder)->bool:
        """
        个人文件下创建文件夹
        Parameters:
            folder (string): 文件夹名称
        Returns:
            创建结果
        """
        result = True
        self._folder_oss = folder+'/'
        exist = self._oss_bucket.object_exists(self._folder_oss)
        if exist is False:
            result = self._oss_bucket.put_object(self._folder_oss,'')
        return result
    def oss_upload_file(self,file_name,local_file)->bool:
        """
        个人文件夹下上传文件
        Parameters:
            file_name (string): 上传上去的文件名称,路径默认为个人根目录,需要特定文件时,需传入路径名称,如:Picfolder/test.png
            localFile (string): 本地文件路径
        Returns:
            上传结果
        """
        result = self._oss_bucket.put_object_from_file(self._folder_oss+file_name,
            local_file)
        return result
    def oss_file_exist(self,file_name)->bool:
        """
        判断个人文件夹下文件是否存在,文件夹请以/结尾
        Parameters:
            file_name (string):上传上去的文件名称,路径默认为个人根目录,需要特定文件时,需传入路径名称,如:Picfolder/test.png
        Returns:
            是否存在
        """
        return self._oss_bucket.object_exists(self._folder_oss+file_name)
    def oss_download_file(self,file_name,local_file):
        """
        下载个人文件
        Parameters:
            file_name (string): 上传上去的文件名称,路径默认为个人根目录,需要特定文件时,需传入路径名称,如:Picfolder/test.png
            local_file (string): 本地路径及文件名
        :return:
        """
        return self._oss_bucket.get_object_to_file(
            self._folder_oss+file_name,
            local_file)
    def oss_list_files(self,folder='')->list:
        """
        获取个人文件夹下文件列表
        Parameters:
            folder (string): 路径默认为个人根目录,需要特定文件时,需传入路径名称,如:Picfolder
        Returns:
            文件列表
        """
        list_file_ret = []
        list_file = oss2.ObjectIterator(
            self._oss_bucket,
            prefix=self._folder_oss+folder)
        for each in list_file:
            list_file_ret.append(each.key)
        return list_file_ret
    def oss_get_file_info(self,file_name):
        """
        获取文件信息
        Parameters:
            file_name (string): 个人文件夹下的文件完整的OSS路径
        Returns:
            文件信息
        """
        return self._oss_bucket.head_object(self._folder_oss+file_name)
    def oss_set_acl(self,file_name,acl):
        """
        设置访问权限
        Parameters:
            file_name (string): 个人文件夹下的文件完整的OSS路径
            scl (int): 1 私有;2 公共读;3 公共读写;默认私有
        Returns:
            设置结果
        """
        # pylint: disable=broad-except
        try:
            if acl == 0:
                return self._oss_bucket.put_object_acl(
                    self._folder_oss+file_name,
                    oss2.OBJECT_ACL_DEFAULT)
            elif acl == 1:
                return self._oss_bucket.put_object_acl(
                    self._folder_oss+file_name,
                    oss2.BUCKET_ACL_PRIVATE)
            elif acl == 2:
                return self._oss_bucket.put_object_acl(
                    self._folder_oss+file_name,
                    oss2.BUCKET_ACL_PUBLIC_READ)
            else:
                raise ValueError('acl should 1 or 2 or 3!')
        except Exception as excp:
            self.logger.error('Exception when set oss acl[%s]',excp)
            [self.logger.error("%s",line) for line in traceback.format_exc().split('\n')]
            return False

    def register_request(self,unit,version,fidx):
        """
        注册对其他单元接口的调用
        Parameters:
            unit (string): 其他单元名,如'pu_fakeif'
            version (int): 其他单元版本号,如1.0.1
            fidx (string): 调用接口,如'example'
        Returns:
            调用句柄
        """
        handle = len(self._reg_request)
        self._reg_request[handle] = {}
        self._reg_request[handle]['unit'] = unit
        self._reg_request[handle]['version'] = version
        self._reg_request[handle]['fidx'] = fidx
        return handle
    @property
    def registerd_request(self)->dict:
        """
        所有对其他单元调用注册
        """
        return self._reg_request

    def register_glb_request(self,unit,version,fidx):
        """
        注册全局系统基础服务的单元接口的调用,全局服务控制各环境,各space的部署及开发者账户等全局信息
        Parameters:
            unit (string): 其他单元名,如'pu_sys_unitcenter'
            version (int): 其他单元版本号,如1.0.1
            fidx (string): 调用接口,如'example'
        Returns:
            调用句柄
        """
        handle = len(self._reg_glb_request)
        self._reg_glb_request[handle] = {}
        self._reg_glb_request[handle]['unit'] = unit
        self._reg_glb_request[handle]['version'] = version
        self._reg_glb_request[handle]['fidx'] = fidx
        return handle+self.GLB_REQUEST_START
    @property
    def registerd_glb_request(self)->dict:
        """
        所有全局单元调用注册
        """
        return self._reg_glb_request

    def set_request_func(self,func):
        """
        框架设置单元调用其他单元的入口函数
        Parameters:
            func (method): 入口函数
        """
        self._request_func = func

    def set_glb_request_func(self,func):
        """
        框架设置单元调用全局基础服务单元的入口函数
        Parameters:
            func (method): 入口函数
        """
        self._request_glb_func = func

    def call_unit(self,handle,index)->tuple:
        """
        调用其他单元接口
        Parameters:
            func (method): 入口函数
        """
        if handle not in self._reg_request:
            raise Exception(f"Handle[{handle}] not registered!")
        return self._request_func(self._reg_request[handle],index)

    def call_glb_unit(self,handle,index)->tuple:
        """
        调用全局基础服务单元接口
        Parameters:
            func (method): 入口函数
        """
        handle -= self.GLB_REQUEST_START
        if handle not in self._reg_glb_request:
            raise Exception(f"Handle[{handle}] not registered!")
        return self._request_glb_func(self._reg_glb_request[handle],index)

    def register_subscription(self,unit,version,fidx,callback):
        """
        注册对其他单元的订阅接口
        Parameters:
            unit (string): 其他单元名,如pu_fakeif
            version (int): 其他单元版本号,如1.0.1
            fidx (string): 调用接口,如'example1'
            callback (function): 回调接口,有数据推送时调用此接口
        Returns:
            订阅句柄
        """
        handle = len(self._reg_subscription)
        self._reg_subscription[handle] = {}
        self._reg_subscription[handle]['unit'] = unit
        self._reg_subscription[handle]['version'] = version
        self._reg_subscription[handle]['fidx'] = fidx
        self._reg_subscription[handle]['callback'] = callback
        return handle
    @property
    def registerd_subscription(self)->dict:
        """
        所有对其他单元订阅注册
        """
        return self._reg_subscription

    def subscribe(self,handle,index):
        """
        订阅其他单元的接口数据更新,订阅需要已经注册过
        Parameters:
            handle (int): 注册过的其他单元接口句柄
        """
        if handle not in self._reg_subscription:
            self.logger.warning("Handle[%d] not registered!",handle)
            return
        self._subscription_func(self._reg_subscription[handle],index)

    def unsubscribe(self,handle,index):
        """
        取消订阅其他单元的接口数据更新,订阅需要已经注册过
        Parameters:
            handle (int): 注册过的其他单元接口句柄
        """
        if handle not in self._reg_subscription:
            self.logger.warning("Handle[%d] not registered!",handle)
            return
        self._unsubscription_func(self._reg_subscription[handle],index)

    def set_subscription_func(self,func):
        """
        框架设置单元订阅其他单元接口的入口函数
        Parameters:
            func (method): 入口函数
        """
        self._subscription_func = func

    def set_unsubscription_func(self,func):
        """
        框架设置单元订阅其他单元接口的入口函数
        Parameters:
            func (method): 入口函数
        """
        self._unsubscription_func = func

class ProcessUnit(UnitBase):
    """
    常驻业务单元基类
    """
    @abstractmethod
    def custom_init(self)->bool:
        """
        各业务单元自定义初始化部分,初始化完成后单元进入正常运行状态,如果出错会自动尝试重新调用此函数再进行初始化直至成功

        Returns:
            True: 初始化完成
            False: 初始化中有错误,会自动重新调用
        """
        return True

class TaskUnit(UnitBase):
    """
    这是说明
    """
    def __init__(self):
        super().__init__()
        self._mutex_current_step = threading.Lock()
        self._dic_current_step = {}
        self.update_step_struct(['initializing','running','done'],'')

    def update_step_struct(self,steps,root_steps)->None:
        """
        设置进度节点，系统默认提供['initializing','running','done']这三个节点
        示例:
        Parameters:
            steps: 进度子节点
            root_steps: 所属的父节点，层级间用';'隔开
        """
        with self._mutex_current_step:
            point_step = self._dic_current_step
            if len(root_steps) > 0:
                roots = root_steps.split(';')
                for level in roots:
                    point_step = point_step['sub_steps'][level]
            point_step['time_start'] = time.time()
            point_step['time_cost'] = 0
            point_step['current_step'] = 0
            point_step['total_step'] = len(steps)
            point_step['current_info'] = steps[0]
            point_step['time_remain'] = 10
            point_step['steps'] = steps
            point_step['sub_steps'] = {}
            for sub_step in steps:
                point_step['sub_steps'][sub_step] = {}
    def _fresh_step(self,step_layer)->None:
        """
        刷新step信息
        Parameters:
            step_layer: 对应的step的列表
        """
        sub_time_remain = 0
        if 'sub_steps' in step_layer:
            for sub in step_layer['sub_steps']:
                if len(step_layer['sub_steps'][sub]) > 0:
                    self._fresh_step(step_layer['sub_steps'][sub])
                if 'time_remain' in step_layer['sub_steps'][sub]:
                    sub_time_remain += step_layer['sub_steps'][sub]['time_remain']
        if 'time_remain' in step_layer:
            step_layer['time_remain'] = max(step_layer['time_remain'],sub_time_remain)
        if 'time_start' in step_layer and 'time_cost' in step_layer:
            step_layer['time_cost'] = time.time()-step_layer['time_start']
    def update_current_step(self,current_step,root_steps)->None:
        """
        更新当前进度
        示例:
        当前进度为  'running;获取期货公司列表;step1',调用方法为:self.update_current_step('step1','runnning;获取期货公司列表')
        Parameters:
            current_step: 当前进度节点
            root_steps: 所属根节点，层级间用';'隔开
        """
        with self._mutex_current_step:
            point_step = self._dic_current_step
            if len(root_steps) > 0:
                roots = root_steps.split(';')
                for level in roots:
                    point_step = point_step['sub_steps'][level]
            cur_index = point_step['steps'].index(current_step)
            if cur_index == 0:
                point_step['time_start'] = time.time()
            else:
                point_step['current_step'] = cur_index
                point_step['current_info'] = current_step
                point_step['time_cost'] = time.time()-point_step['time_start']
                avgtime_step = point_step['time_cost']/cur_index
                point_step['time_remain'] = avgtime_step*(point_step['total_step']-cur_index-1)
                self._fresh_step(self._dic_current_step)
            self.logger.info('%s',self._dic_current_step)

    def req_get(self,url, params=None, **kwargs):
        """
        封装reuest.get,调用频率控制,缓存,错误打印,ip池等
        """
        resp = requests.get(url, params=params, **kwargs)
        time.sleep(random.uniform(0.5,3))
        if resp.status_code != 200:
            self.logger.error('Get request [%s] failed, code[%d], info[%s]',
            url,resp.status_code,resp.text)
        return resp

    def req_post(self,url, data=None, json=None, **kwargs):
        """
        封装reuest.post,调用频率控制,缓存,错误打印,ip池等
        """
        resp = requests.post(url, data=data, json=json, **kwargs)
        time.sleep(random.uniform(0.5,3))
        if resp.status_code != 200:
            self.logger.error('Post request [%s] failed, code[%d], info[%s]',
            url,resp.status_code,resp.text)
        return resp

    @abstractmethod
    def run(self)->tuple:
        """
        执行task unit任务

        Returns:
            code: 执行结果代码
            info: 执行结果信息
            content: 执行结果
        """
        return (0,'','')

class TradeStrategyUnit(UnitBase):
    """
    这是说明
    """
    def __init__(self):
        super().__init__()
        self._user = ''

    def hello(self):
        """
        这是说明
        """
        self._user = 'username'

    def hello2(self):
        """
        这是说明
        """
        self._user = 'username'
